import numpy as np
import valclust
import sys

def test_get_members():
    X = np.array([1,1,1,1,1,1,1,1,1])
    a = np.array([1,1,1,2,3,1,1,2,3])

    cobj = valclust.cluster.Cluster(X, a)

    a1 = cobj._get_members(1)
    
    assert(a1.shape[0] == 5)

    a1_true = [0,1,2,5,6]
    for i in range(5):
	assert(a1[i] == a1_true[i])


def test_get_members_string():
    X = np.array(['TATT', 'TTTA', 'GACG', 'GCGT', 'ATTA', 'TAAT', 'GCGG'])
    y = np.array(['c1', 'c1', 'c2', 'c2', 'c1', 'c1', 'c2'])

    cobj = valclust.cluster.Cluster(X, y)
    a1 = cobj._get_members('c1')
    a2 = cobj._get_members('c2')

    assert (a1.shape[0] == 4)
    assert (a2.shape[0] == 3)

    a1_true = [0,1,4,5]
    for i in range(4):
       assert(a1[i] == a1_true[i])

    a2_true = [2,3,6]
    for i in range(3):
       assert(a2[i] == a2_true[i])


def test_get_non_members():
    X = np.array([1,1,1,1,1,1,1,1,1])
    a = np.array([1,1,1,2,3,1,1,2,3])

    cobj = valclust.cluster.Cluster(X, a)

    anot1 = cobj._get_non_members(1)

    assert(anot1.shape[0] == 4)

    anot1_true = [3,4,7,8]
    for i in range(4):
        assert(anot1[i] == anot1_true[i])


def test_get_num_singletons():
    X = np.array([1,1,1,1,1,1,1,1,1,1,1,1])
    a = np.array([1,1,0,2,3,0,1,0,4,0,1,2])

    cobj = valclust.cluster.Cluster(X, a)
    nsing = cobj._num_singletons()
    assert (nsing == 2)

    a = np.arange(X.shape[0])
    cobj = valclust.cluster.Cluster(X, a)
    nsing = cobj._num_singletons()
    assert (nsing == X.shape[0])

    cobj.summary()



import numpy as np
import sys
from valclust.InternalValidity import PairwiseDistanceSampler as pds


def test_intra_sampling():

    np.random.seed(1234)

    d1 = np.random.multivariate_normal(mean=[-1,1], cov=[[1, 0],[0, 1]], size=3)
    d2 = np.random.multivariate_normal(mean=[2,-2], cov=[[1, 0],[0, 1]], size=4)

    d = np.vstack((d1, d2))
    y = np.array([1,1,1, 2,2,2,2])

    sam_obj = pds(d, y)

    ds1 = sam_obj.intra_sampler(1, size=None)
    assert(ds1.shape[0]) == 3

    ds2 = sam_obj.intra_sampler(2, size=None)
    assert(ds2.shape[0]) == 6


    #def test_intra_Sampling_string():
    d = np.array(["TTTA", "TGTT", "TAAT", "GTTA", "GGCG"])
    y = np.array(["c1", "c1", "c1", "c1", "c2"])

    sam_obj = pds(d, y)

    ds3 = sam_obj.intra_sampler("c1", size=None, method="editdist")
    assert(ds3.shape[0]) == 6


def test_inter_sampling():

    np.random.seed(1234)

    d1 = np.random.multivariate_normal(mean=[-1,1], cov=[[1, 0],[0, 1]], size=3)
    d2 = np.random.multivariate_normal(mean=[2,-2], cov=[[1, 0],[0, 1]], size=4)

    d = np.vstack((d1, d2))
    y = np.array([1,1,1, 2,2,2,2])

    sam_obj = pds(d, y)

    ds1 = sam_obj.inter_sampler(1, size=None)
    assert(ds1.shape[0]) == 12






import numpy as np
import valclust
import valclust.external as exv
import sys

def test_purity():
    y = np.array([1,1,1,2,2,2,2,3])
    g = np.array([5,5,5,7,7,9,9,5])

    obj = exv.CompareCluster(X=None, y=y, g=g)


    r1 = obj.clusterPurity(1)
    assert (r1 == 1.0)

    r2 = obj.clusterPurity(2)
    assert (r2 == 0.5)

    r3 = obj.clusterPurity(3)
    assert (r3 == 1.0)

    tot,wpurity = obj.totalPurity()
    assert (tot == 5/6.0 and wpurity==3/4.0)

def test_NMI():
    y = np.array([1,1,1,2,2,2,2,3])
    g = np.array([5,5,5,7,7,5,5,7])

    obj = exv.CompareCluster(X=None, y=y, g=y)

    nmi = obj.normalizedMutualInfo() ## Comparre with itself
    sys.stderr.write("NMI : %f"%(nmi))
    assert (nmi == 1.0)

    obj = exv.CompareCluster(X=None, y=y, g=g)
    nmi = obj.normalizedMutualInfo()
    sys.stderr.write("NMI : %f"%(nmi))
    assert(int(nmi*10000) == 3851)

def test_contingency():
    y = np.array([1,1,1,1,1,1,2,2,2,2,2,2,3,3,3,3,3])
    g = np.array([4,4,4,4,5,4,4,5,5,5,7,5,4,4,7,7,7])

    obj = exv.CompareCluster(X=None, y=y, g=g)
    obj._contingency()

    #sys.stderr.write("TP:  %f\tFP:  %f\n"%(obj.tp, obj.fp))
    #sys.stderr.write("FN:  %f\tTN:  %f\n"%(obj.fn, obj.tn))
    assert(obj.tp == 20)
    assert(obj.fp == 20)
    assert(obj.fn == 24)
    assert(obj.tn == 72)

    obj = exv.CompareCluster(X=None, y=y, g=y)
    obj._contingency()
    assert(obj.tp == 40)
    assert(obj.fp == 0)
    assert(obj.fn == 0)
    assert(obj.tn == 96)

def test_MCC():
    y = np.array([1,1,1,1,1,1,2,2,2,2,2,2,3,3,3,3,3])
    g = np.array([4,4,4,4,5,4,4,5,5,5,7,5,4,4,7,7,7])

    obj = exv.CompareCluster(X=None, y=y, g=g)
    res = obj.MCC()

    assert(int(res*10000) == 2434)



# Vahid Mirjalili
# Clustering Validation and Analysis

import numpy as np
import scipy.special as sp
import pyprind
import sys
from .cluster import Cluster
import editdist


class CompareCluster(Cluster):

    def __init__(self, X, y, g):
        assert (y.shape == g.shape)
        self.set_data(X, y)
	self.g = g

    def clusterPurity(self, cinx):
	""" Compute the purity of an individual cluster cinx 
	with respect to a given clustering solution 
	g (which could be the ground-truth partitioning).
	"""
	cmembers = self._get_members(cinx)
	memg = self.g[cmembers]
	sdict = {}
	for k in memg:
	    if k in sdict.keys():
	        sdict[k] += 1
	    else:
		sdict[k] = 1

	largest = np.max(sdict.values())
	return (largest / float(memg.shape[0]))


    def totalPurity(self):
        """ Compute the total purity measure w.r.t. g partitioning
	"""
	res, n = 0.0, 0
	wres, wsum = 0.0, 0.0
	for k in np.unique(self.y):
	    p = self.clusterPurity(k)
	    res += p
	    n += 1
	    ksize = self.clsize[self.clsize[:,0] == k,1]
	    wres += p*ksize
	    wsum += ksize
	return (res/n, wres/wsum)


    def normalizedMutualInfo(self):
	""" Compute Normalized Mutual Information (NMI) 
		w.r.t. ground truth partitioning.
	"""
	n = self.n

	y_uniq = self.unique()

	gc = Cluster(X=None, y=self.g)
	g_uniq = gc.unique()

	y_csize = self.cluster_sizes()
	g_csize = gc.cluster_sizes()


	sval = 0.0
	for i in y_uniq:
	    imembers = self._get_members(i)

	    isize = y_csize[y_csize[:,0]==i,1]
	    gsub = self.g[imembers]

	    for j in np.unique(gsub):
		size_ij = np.sum(gsub == j)
		jsize = g_csize[g_csize[:,0]==j,1]
		#print("%d %d \t %d %d --> %d  %f" %(i, isize, j, jsize, size_ij, size_ij * n/float(isize * jsize)))
		sval += size_ij/float(n) * np.log(size_ij * n/float(isize * jsize))

	return(2.0 * sval / (self.entropy() + gc.entropy()))

    def MCC(self):
	""" Compute Matthew's Correlation Coefficient:
		MCC = (TP*TN - FP*FN) / sqrt((TP+FP)*(TP+FN)*(TN+FP)*(TN+FN))
	"""
	self._contingency()
	tp,fp,tn,fn = self.tp, self.fp, self.tn, self.fn
	return ((tp*tn - fp*fn)/np.sqrt((tp+fp)*(tp+fn)*(tn+fp)*(tn+fn)))

    def _contingency(self):
	""" Compute TP, FP, TN, and FN
		for any pair of points.
	"""
	tp_fp, tp = 0.0, 0.0
	tn_fn, fn = 0.0, 0.0
	for i,v1 in enumerate(self.clsize[:,:]):
	    c1,n1 = v1
	    imemb1 = self._get_members(c1)
	    gsub1 = self.g[imemb1]
	    if n1 >=2:
		tp_fp += sp.binom(n1, 2)
		for j in np.unique(gsub1):
		    size_ij = np.sum(gsub1 == j)
		    if size_ij >= 2:
			tp += sp.binom(size_ij, 2)
		    size_j = np.sum(self.g == j)
		    fn += size_ij * (size_j - size_ij)
		    #print("%d %d %d   %d\t\t%d"%(j, size_j, size_ij, size_ij*(size_j-size_ij), fn))

	    if (i < self.clsize.shape[0] - 1): 
		tn_fn += n1 * np.sum(self.clsize[i+1:,1])


	self.tp = tp
	self.fp = tp_fp - tp

	fn = fn / 2
	self.tn = tn_fn - fn
	self.fn = fn

	#sys.stderr.write("TP: %.0f  FP: %.0f \t TN: %.0f  FN: %.0f\n"%(self.tp,self.fp,self.tn,self.fn))
	return (None)



# Vahid Mirjalili
# Clustering Validation and Analysis

import numpy as np
import pyprind
import sys
from ..cluster import Cluster
import editdist

def cal_distance(x1, x2, method='euclidean'):
    if method == 'euclidean':
        diff = x1 - x2
        sq_diff = diff * diff
        return(np.sqrt(np.sum(sq_diff)))
    elif method == 'editdist':
	return(editdist.distance(x1, x2))


class PairwiseDistanceSampler(Cluster):
    """
    Class for sampling intra and inter cluster pairwise distances

    Class Methods:

    intra_sampler()
    inter_sampler()

    """

    def __init__(self, X, y):
        self.set_data(X, y)

    def intra_sampler(self, cinx, size=None, method='euclidean', return_ind=False, outfile=None):
	"""
	Sampling intra-cluster pairwise distances.

	Arguments: 
		X (feature vectors/data)
		y (cluster memberships)
		cinx (distinct cluster index) to run sampling for a particular cluster
		size (amount of sampling), if None, all exclusive pairwise distances are computed.
		method: distance method, such as 'euclidean', 'mahalanobis', ...

	"""

	X_clust = self.X[self._get_members(cinx)]

        nsize = X_clust.shape[0]

	totpw = size
	if totpw is None:
	   totpw = nsize*(nsize-1)/2
	sys.stderr.write("##\tCluster %s  --> Size: %d\n" %(cinx, nsize))


	rand_pair_ind = np.random.randint(low=0, high=nsize, size=(totpw,2))

	if outfile is None:
	    dary = np.zeros(shape=totpw, dtype='float')
	    n = 0
	    for i,j in rand_pair_ind:
		dary[n] = cal_distance(X_clust[i], X_clust[j], method=method)
        	n += 1
	    return(dary)
	else:
	    with open(outfile, 'w') as fp:
		for i,j in rand_pair_ind:
		   dist_ij = cal_distance(X_clust[i], X_clust[j], method=method)
		   fp.write("%-d %-d %.4f\n" %(i, j, dist_ij))

	    return(None)




    def inter_sampler(self, cinx, size=None, method='euclidean', return_ind=False, outfile=None):
	"""
	Sampling inter-cluster pairwise distances 
	for a particular cluster (cinx) with respect to all other clusters.


	"""
        memb_indx = self._get_members(cinx)
	rest_indx = self._get_non_members(cinx)

        nsize_memb = memb_indx.shape[0]
	nsize_rest = rest_indx.shape[0]

	totpw = size
	if totpw is None:
	    totpw = nsize_memb * nsize_rest

	sys.stderr.write("##\tCluster %s w.r.t. rest --> Size: %d x %d  Total pairs: %d\n" 
		%(cinx, nsize_memb, nsize_rest, totpw))

	rand_memb_ind = memb_indx[np.random.randint(low=0, high=nsize_memb, size=totpw)]
	rand_rest_ind = rest_indx[np.random.randint(low=0, high=nsize_rest, size=totpw)]

	
	if outfile is None:
	    # No file is given, so save the output in a numpy array
	    dary = np.zeros(shape=totpw, dtype='float')
            n = 0 
            for i,j in zip(rand_memb_ind, rand_rest_ind):
                dary[n] = self.distance(i, j) #, method=method)
                n += 1
	    return(dary)
	
	else:
	    # Save the output in the given file
	    with open(outfile, 'w') as fp:
	       for i,j in zip(rand_memb_ind, rand_rest_ind):
	          dist_ij = self.distance(i, j, method=method)
	          fp.write("%-d %-d %.4f\n" %(i, j, dist_ij))

	    return(None)



# Vahid Mirjalili 12/31/2014
# Clustering Validation and Analysis
# Submodules for sampling intra and inter cluster distances


from ..cluster import Cluster

from .pdist_sampling import PairwiseDistanceSampler




import numpy as np

class Cluster(object):
    """
    Class to find distinct clusters.

    Class Metods:

	n_distnict(self, y): the number of distinct clusters
	cluster_sizes(self, y): the sizes of each cluster in a numpy ndarray
	get_members(self, y, cinx): the indeces of members of a particular cluster cinx

    """

    def __init__(self, X=None, y=None):
	self.set_data(X, y)

    def set_data(self, X, y):
        self.X = X
        self.y = y
        self.n = y.shape[0]
	self.clsize = self.cluster_sizes()

    def n_distinct(self):
	"""
	Returns  the number of distinct clusters.
	"""
        return(np.unique(self.y).shape[0])

    def unique(self):
	return(np.unique(self.y))

    def cluster_sizes(self):
	"""
	Returns the sizes of each cluster in a numpy ndarray.
	"""
	tup = np.unique(self.y, return_counts = True)
	return(np.asarray(tup).T)

    def _get_members(self, cinx):
	"""
	Returns a numpy array for the indecies of all 
	members for a specific cluster indexed by cinx.
	
	"""
	allindx = np.arange(0, self.n)
	return (allindx[self.y == cinx])
        

    def _get_non_members(self, cinx):
	"""
	Returns a numpy array for the indecies of all
	data points that DO NOT belong to cluster cinx.

	"""
	allindx = np.arange(0, self.n)
	return (allindx[self.y != cinx])

    def distance(self, i, j, method='euclidean'):
        x1 = self.X[i]
        x2 = self.X[j]
        if method == 'euclidean':
            diff = x1 - x2
            sq_diff = diff * diff
            return(np.sqrt(np.sum(sq_diff)))
        elif method == 'editdist':
            return(editdist.distance(x1, x2))

    def _num_singletons(self):
	"""Finding the number of singletons
	   specified by indicator [default=-1]
	"""
        return (np.sum(self.clsize[:,1] == 1))

    def entropy(self):
	""" Computing the entropy of a clustering.
	"""
	s = 0.0
	nt = float(self.n)
	for k in self.clsize[:,1]:
	    s -= k/nt * np.log(k/nt)
	return(s)

    def summary(self, sing_indicator=-1):
        """ Gives a summary statistics on clusters.
	"""
	print ("Number of singletons: %d"%(self._num_singletons()))
	print ("Number of clusters: %d"%(self.n_distinct()))

	return (None)



import numpy as np
import scipy
import scipy.spatial
import sklearn.cluster


def DBSCAN(data, eps, min_pts):
    """ Clustering data using DBSCAN algorithm.
    """
    
    #kd-tree for quick nearest-neighbor lookup
    tree = scipy.spatial.cKDTree(data)
    neighbors = tree.query_ball_point(data, eps)
    clusters = [] # list of (set, set)'s, to distinguish 
                  # core/reachable points
    visited = set()
    for i in xrange(len(data)):
        if i in visited: 
            continue
        visited.add(i)
        if len(neighbors[i]) >= min_pts:
            clusters.append(({i}, set())) # core
            to_merge_in_cluster = set(neighbors[i])
            while to_merge_in_cluster:
                j = to_merge_in_cluster.pop()
                if j not in visited:
                    visited.add(j)
                    if len(neighbors[j]) >= min_pts:
                        to_merge_in_cluster |= set(neighbors[j])
                if not any([j in c[0] | c[1] for c in clusters]):
                    if len(neighbors[j]) >= min_pts:
                        clusters[-1][0].add(j) # core
                    else:
                        clusters[-1][1].add(j) # reachable
    return np.array(clusters)


import networkx as nx

colors = 'rbgycm' * 10

def ConnectedComponents(data, eps):

   tree = scipy.spatial.cKDTree(data)
   neighbors = tree.query_ball_point(data, eps)

   g = nx.Graph()
   g.add_nodes_from(xrange(len(data)))
   nx.draw_networkx(g, data, with_labels=False, node_size=5, \
                 node_color=[.5] * 3, width=.5)   
   g = nx.Graph()
   for i in xrange(len(data)):
       if len(neighbors[i]) >= 10:
           g.add_edges_from([(i, j) for j in neighbors[i]])

   for i, comp in enumerate(nx.connected_component_subgraphs(g)):
        nx.draw_networkx(comp, data, with_labels=False, node_size=5, \
                     node_color=[.5] * 3, edge_color=colors[i], width=.5)




def main():

    np.random.seed(seed=1234)

    m1 = np.array([-2,-2])
    m2 = np.array([0,1.5])
    m3 = np.array([1.7,-1.8])
    S1 = np.array([[0.4,0.1],[0.1,0.4]])
    S2 = np.array([[0.5,-0.05],[-0.05,0.5]])
    S3 = np.array([[0.56,0],[0,0.56]])
    d1 = np.random.multivariate_normal(mean=m1, cov=S1, size=1000)
    d2 = np.random.multivariate_normal(mean=m2, cov=S2, size=1000)
    d3 = np.random.multivariate_normal(mean=m3, cov=S3, size=1000) 

    d = np.concatenate([d1, d2, d3])
    print(d.shape)

    #np.savetxt('zzz/data.3clusters.dat', d)


    import time
   
    start = time.time()
    cl = DBSCAN(d, eps=0.2, min_pts=20)
    end = time.time()

    print(end-start, cl.shape)
    for i, (core, neighb) in enumerate(cl):
	print (i, len(core), len(neighb), len(core) + len(neighb))


    dbs = sklearn.cluster.DBSCAN(eps=0.2, min_samples=20, metric='euclidean', algorithm='kd_tree', leaf_size=100)
    start = time.time()
    cl2 = dbs.fit_predict(d)
    end = time.time()
    print(end-start, cl2.shape)
    for i in np.unique(cl2):
        print (i, np.sum(cl2 == i))

    dbs = sklearn.cluster.DBSCAN(eps=0.2, min_samples=20, metric='euclidean', algorithm='brute')
    start = time.time()
    cl2 = dbs.fit_predict(d)
    end = time.time()
    print(end-start, cl2.shape)

    #ConnectedComponents(d, 0.2)

if __name__ == "__main__":
    main()





from distutils.core import setup


files = ["things/*"]

setup(name = "valclust",
    version = "0.1.6",
    description = "Clustering Validation and Analysis",
    author = "Vahid Mirjalili",
    author_email = "vmirjalily@gmail.com",
    url = "https://github.com/mirjalil/valclust",

    packages = ['valclust', 'valclust.InternalValidity',
	'doc', 'tests'],

    #package *needs* these files.
    package_data = {'valclust':[]},
    classifiers = [
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
	"Intended Audience :: Information Technology",
	"Operating System :: OS Independent",
    ],

    scripts = [],
    long_description = """

A collection of tools for validation and analysis of clustering results.

Currently Supported:  
   * Summary Statistics of Clustering: singletons, cluster sizes
   * Internal Indecies: Inter \& Intra Pairwise Distance Sampling
   * External Validation:  Purity, NMI, MCC, Binary Contingency


 Contact
=============

email: vmirjalily@gmail.com
Twitter: https://twitter.com/vmirly

URL for this project: https://github.com/mirjalil/valclust

""", 
    #classifiers = [],
    license='GPLv3',
    platforms='any',
) 


